import hashlib
from pathlib import Path
from typing import List, Tuple

import chromadb
import ollama

from open_terminalui._models import Chat, Message


class MemoryManager:
    """Manages chat message vector embeddings using ChromaDB"""

    def __init__(self, storage_path: str | None = None):
        """Initialize the document manager with ChromaDB client"""
        if storage_path is None:
            # Default to ~/.open-terminalui/chroma_db
            app_dir = Path.home() / ".open-terminalui"
            app_dir.mkdir(exist_ok=True)
            storage_path = str(app_dir / "chroma_db")

        self.storage_path = storage_path
        self.client = chromadb.PersistentClient(path=storage_path)

        # Get or create the documents collection
        self.collection = self.client.get_or_create_collection(
            name="chat-message-summaries",
            metadata={"description": "Chat message summaries with embeddings"},
        )

    def _get_chat_message_hash(self, chat_id: int, message_index: int) -> str:
        """Generate a hash for the file to use as unique identifier"""
        return hashlib.md5(f"{chat_id}-{message_index}".encode()).hexdigest()

    def _summarize_message(self, message: Message, min_length: int = 200) -> str | None:
        """
        Summarize a message if it exceeds min_length, otherwise return original content.

        Args:
            message: The message to summarize
            min_length: Minimum character length to trigger summarization (default: 200)

        Returns:
            Summary if message is long enough, otherwise original content
        """
        # If message is short, return it as-is without summarization
        if len(message.content) < min_length:
            return message.content

        prompt = f"""Help me build your memory by summarizing this message from the user into just the facts:

Message: {message.content}

Provide only the summary without labels:
"""

        ollama_request = [{"role": "user", "content": prompt}]

        try:
            ollama_response = ollama.chat(
                model="llama3.2", messages=ollama_request, stream=False
            )

            message_summary = ollama_response.message.content

            if message_summary is not None:
                return message_summary

        except Exception as e:
            raise Exception(e)

    def save_chat(self, chat: Chat):
        """
        Save chat messages to the vector store with embeddings.

        Processes each user and assistant message in the chat, summarizes them if needed,
        and stores them in ChromaDB for semantic search. Skips messages that have already
        been saved.

        Args:
            chat: The chat object containing messages to save
        """
        if chat.id is None:
            return

        chat_id = chat.id
        for message_index, message in enumerate(chat.messages):
            # Only index user and assistant messages (skip log messages)
            if message.role not in ("user", "assistant"):
                continue

            chat_message_hash = self._get_chat_message_hash(chat_id, message_index)
            existing = self.collection.get(
                where={"chat_message_hash": chat_message_hash}
            )

            # Check if message has already been saved
            if existing and existing["ids"]:
                return None

            message_summary = self._summarize_message(message)

            if message_summary is None:
                continue

            metadata = {
                "chat_id": chat_id,
                "message_index": message_index,
                "chat_message_hash": chat_message_hash,
            }

            self.collection.add(
                ids=[chat_message_hash],
                documents=[message_summary],
                metadatas=[metadata],
            )

    def delete_chat(self, chat_id: int):
        """Delete all message summaries associated with a chat"""
        # Query for all documents with this chat_id
        try:
            results = self.collection.get(where={"chat_id": chat_id})

            # Delete all matching documents
            if results and results["ids"]:
                self.collection.delete(ids=results["ids"])

        except Exception as e:
            raise Exception(f"Failed to delete chat {chat_id}: {e}")

    def list_chat_summaries(self) -> List[Tuple[int, int, str]]:
        """
        List all chat summaries in the vector store

        Returns:
            List of tuples: (chat_id, message_count, chat_message_hashes)
        """
        try:
            # Get all chat summaries
            results = self.collection.get()

            if not results["metadatas"]:
                return []

            # Group by chat_id to get unique chats
            chats_by_id = {}
            for metadata in results["metadatas"]:
                chat_id = metadata["chat_id"]
                if chat_id not in chats_by_id:
                    chats_by_id[chat_id] = {"message_count": 0, "hashes": []}
                chats_by_id[chat_id]["message_count"] += 1
                chats_by_id[chat_id]["hashes"].append(metadata["chat_message_hash"])

            # Convert to list of tuples
            return [
                (chat_id, chat_data["message_count"], ", ".join(chat_data["hashes"]))
                for chat_id, chat_data in chats_by_id.items()
            ]

        except Exception as _:
            return []

    def search_chat_summaries(
        self, query: str, top_k: int = 5
    ) -> List[Tuple[str, float]]:
        """
        Search chat message summaries using semantic similarity.

        Performs a vector similarity search across all stored chat message summaries
        and returns the most relevant results.

        Args:
            query: The search query text
            top_k: Maximum number of results to return (default: 5)

        Returns:
            List of tuples containing (document_text, similarity_score) where
            similarity_score is between 0 and 1, with 1 being most similar

        Raises:
            Exception: If the search operation fails
        """
        try:
            results = self.collection.query(query_texts=[query], n_results=top_k)

            if not results["documents"] or not results["documents"][0]:
                return []

            chunks = []
            documents = results["documents"][0]
            distances = results["distances"][0] if results["distances"] else []

            for i, doc in enumerate(documents):
                # ChromaDB returns distances, lower is better
                # Convert to similarity score (1 - normalized_distance)
                distance = distances[i] if distances else 0
                similarity = max(0, 1 - distance)

                chunks.append((doc, similarity))

            return chunks

        except Exception as e:
            raise Exception(e)



from textual.app import ComposeResult
from textual.widgets import (
    Label,
    ListItem,
)


class ChatListItem(ListItem):
    """A custom list item for displaying chat titles"""

    def __init__(self, chat_id: int, title: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.chat_id = chat_id
        self.chat_title = title

    def compose(self) -> ComposeResult:
        yield Label(self.chat_title)



import json
import sqlite3
from datetime import datetime
from pathlib import Path

from ._models import Chat, Message


class ChatManager:
    """Manages chat persistence using SQLite database"""

    def __init__(self, db_path: str | None = None):
        """
        Initialize the chat manager with SQLite database.

        Args:
            db_path: Path to the SQLite database file. If None, defaults to
                    ~/.open-terminalui/chats.db
        """
        if db_path is None:
            # Default to ~/.open-terminalui/chats.db
            app_dir = Path.home() / ".open-terminalui"
            app_dir.mkdir(exist_ok=True)
            db_path = str(app_dir / "chats.db")

        self.db_path = db_path
        self._init_db()

    def _init_db(self):
        """
        Initialize the database schema.

        Creates the chats table if it doesn't exist with columns for id, title,
        messages_json, created_at, and updated_at.
        """
        with sqlite3.connect(self.db_path) as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS chats (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    title TEXT NOT NULL,
                    messages_json TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL
                )
            """)
            conn.commit()

    def create_chat(self, title: str | None = None) -> Chat:
        """
        Create a new chat and save it to the database.

        Args:
            title: The title for the chat. If None, generates a default title
                  with the current date and time (e.g., "Chat 2025-12-11 14:30")

        Returns:
            A new Chat object with an assigned ID and empty message list
        """
        if title is None:
            title = f"Chat {datetime.now().strftime('%Y-%m-%d %H:%M')}"

        now = datetime.now()
        messages_json = json.dumps([])

        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.execute(
                "INSERT INTO chats (title, messages_json, created_at, updated_at) VALUES (?, ?, ?, ?)",
                (title, messages_json, now.isoformat(), now.isoformat()),
            )
            conn.commit()
            chat_id = cursor.lastrowid

        return Chat(
            id=chat_id, title=title, messages=[], created_at=now, updated_at=now
        )

    def save_chat(self, chat: Chat):
        """
        Save or update a chat in the database.

        If the chat has no ID (newly created), inserts it as a new record.
        If the chat has an ID (existing), updates the existing record.
        Automatically updates the updated_at timestamp.

        Args:
            chat: The Chat object to save. The chat.id will be set if it's a new chat.
        """
        messages_json = json.dumps([msg.to_dict() for msg in chat.messages])
        chat.updated_at = datetime.now()

        with sqlite3.connect(self.db_path) as conn:
            if chat.id is None:
                # Chat doesn't exist in DB yet, insert it
                cursor = conn.execute(
                    "INSERT INTO chats (title, messages_json, created_at, updated_at) VALUES (?, ?, ?, ?)",
                    (
                        chat.title,
                        messages_json,
                        chat.created_at.isoformat(),
                        chat.updated_at.isoformat(),
                    ),
                )
                chat.id = cursor.lastrowid
            else:
                # Chat exists, update it
                conn.execute(
                    "UPDATE chats SET title = ?, messages_json = ?, updated_at = ? WHERE id = ?",
                    (chat.title, messages_json, chat.updated_at.isoformat(), chat.id),
                )
            conn.commit()

    def load_chat(self, chat_id: int) -> Chat | None:
        """
        Load a chat from the database by its ID.

        Args:
            chat_id: The unique identifier of the chat to load

        Returns:
            The Chat object if found, None if no chat exists with the given ID
        """
        with sqlite3.connect(self.db_path) as conn:
            conn.row_factory = sqlite3.Row
            cursor = conn.execute("SELECT * FROM chats WHERE id = ?", (chat_id,))
            row = cursor.fetchone()

        if row is None:
            return None

        messages_data = json.loads(row["messages_json"])
        messages = [
            Message(role=msg["role"], content=msg["content"]) for msg in messages_data
        ]

        return Chat(
            id=row["id"],
            title=row["title"],
            messages=messages,
            created_at=datetime.fromisoformat(row["created_at"]),
            updated_at=datetime.fromisoformat(row["updated_at"]),
        )

    def list_chats(self) -> list[Chat]:
        """
        List all chats from the database.

        Returns:
            List of all Chat objects ordered by updated_at timestamp in descending order
            (most recently updated first). Returns empty list if no chats exist.
        """
        with sqlite3.connect(self.db_path) as conn:
            conn.row_factory = sqlite3.Row
            cursor = conn.execute(
                "SELECT id, messages_json, title, created_at, updated_at FROM chats ORDER BY updated_at DESC"
            )
            rows = cursor.fetchall()
        # messages_json = json.dumps([msg.to_dict() for msg in chat.messages])

        chats = []
        for row in rows:
            messages_data = json.loads(row["messages_json"])
            messages = [
                Message(role=msg["role"], content=msg["content"])
                for msg in messages_data
            ]
            chats.append(
                Chat(
                    row["id"],
                    row["title"],
                    messages,
                    datetime.fromisoformat(row["updated_at"]),
                    datetime.fromisoformat(row["updated_at"]),
                )
            )

        return chats

    def delete_chat(self, chat_id: int):
        """
        Delete a chat from the database.

        Args:
            chat_id: The unique identifier of the chat to delete
        """
        with sqlite3.connect(self.db_path) as conn:
            conn.execute("DELETE FROM chats WHERE id = ?", (chat_id,))
            conn.commit()



from open_terminalui.memory_manager import MemoryManager


def memory_search(
    memory_manager: MemoryManager, query: str, max_results: int = 5
) -> str:
    """Perform document search and return formatted results"""
    try:
        # Search vector database
        results = memory_manager.search_chat_summaries(query=query, top_k=max_results)

        # Format results for LLM context
        formatted_results = ""
        for result in results:
            formatted_results += f"Content: {result[0]}\n"
            formatted_results += f"Similarity Score: {result[1]}\n\n"

        return formatted_results
    except Exception as e:
        return f"Vector search error: {str(e)}"



from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    DataTable,
    Input,
    Label,
    Static,
)

from open_terminalui.document_manager import DocumentManager


class DocumentManagerScreen(ModalScreen):
    """Modal screen for managing documents"""

    def __init__(self, doc_manager: DocumentManager, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doc_manager = doc_manager

    def compose(self) -> ComposeResult:
        with Vertical(id="document_dialog"):
            yield Label("Manage Documents", id="document_title")
            yield Static(id="status_indicator")
            with Horizontal(id="document_input_container"):
                yield Input(
                    placeholder="Enter PDF file path...",
                    id="document_path_input",
                )
                yield Button("Add", id="add_document_btn", variant="primary")
            yield DataTable(id="document_table")
            with Horizontal(id="document_button_container"):
                yield Button(
                    "Remove Selected", id="remove_document_btn", variant="error"
                )
                yield Button("Close", id="close_dialog_btn", variant="default")

    def on_mount(self) -> None:
        # Configure data table
        table = self.query_one("#document_table", DataTable)
        table.cursor_type = "row"
        table.add_columns("Name", "Path", "Chunks")

        # Load table rows
        self._refresh_table()

    def _refresh_table(self) -> None:
        # Select table widget
        table = self.query_one("#document_table", DataTable)

        # Clear existing rows
        table.clear()

        # Get documents
        documents = self.doc_manager.list_documents()

        # Populate table
        for document in documents:
            table.add_row(document[1], document[0], document[2], key=str(document[0]))

    @on(Button.Pressed, "#add_document_btn")
    def handle_add_document(self) -> None:
        # Select input widget
        input_widget = self.query_one("#document_path_input", Input)

        # Get file_path
        file_path: str = input_widget.value.strip()

        # Clear input widget
        input_widget.clear()

        # Process document
        self.process_document(file_path)

    @work(exclusive=True, thread=True)
    def process_document(self, file_path: str) -> None:
        # Select widgets
        status_widget = self.query_one("#status_indicator", Static)
        input_widget = self.query_one("#document_path_input", Input)
        button_widget = self.query_one("#add_document_btn", Button)

        # Disable button and show processing status
        self.app.call_from_thread(setattr, button_widget, "disabled", True)
        self.app.call_from_thread(status_widget.update, "Processing...")

        # Process document
        success, message = self.doc_manager.add_document(file_path)

        # Update status and re-enable button
        self.app.call_from_thread(status_widget.update, message)
        self.app.call_from_thread(setattr, button_widget, "disabled", False)

        # If successful, clear input widget and update table
        if success:
            self.app.call_from_thread(input_widget.clear)
            self.app.call_from_thread(self._refresh_table)

    @on(Button.Pressed, "#remove_document_btn")
    def handle_remove_document(self) -> None:
        # Select widgets
        table = self.query_one("#document_table", DataTable)
        status_widget = self.query_one("#status_indicator", Static)

        # Get the selected row's file path
        if table.cursor_row is not None:
            row = table.get_row_at(table.cursor_row)
            file_path = row[1]

            # Remove the document
            success, message = self.doc_manager.remove_document(file_path)
            status_widget.update(message)

            # Refresh table if successful
            if success:
                self._refresh_table()

    @on(Button.Pressed, "#close_dialog_btn")
    def handle_close_dialog(self) -> None:
        self.app.pop_screen()



from dataclasses import dataclass
from datetime import datetime


@dataclass
class Message:
    role: str  # "user" or "assistant"
    content: str

    def to_dict(self) -> dict:
        return {"role": self.role, "content": self.content}


@dataclass
class Chat:
    id: int | None
    title: str
    messages: list[Message]
    created_at: datetime
    updated_at: datetime

    def to_ollama_messages(self) -> list[dict]:
        """Convert messages to Ollama API format, excluding log messages"""
        return [
            msg.to_dict()
            for msg in self.messages
            if msg.role in ("user", "assistant", "system")
        ]

    @staticmethod
    def create_unsaved(title: str | None = None) -> "Chat":
        """Create a new chat that hasn't been saved to the database yet"""
        if title is None:
            title = f"Chat {datetime.now().strftime('%Y-%m-%d %H:%M')}"

        now = datetime.now()
        return Chat(id=None, title=title, messages=[], created_at=now, updated_at=now)



from ddgs import DDGS


def web_search(query: str, max_results: int = 5) -> str:
    """Perform web search and return formatted results"""
    try:
        results = list(DDGS().text(query, max_results=max_results))
        if not results:
            return "No search results found."

        # Format results for LLM context
        formatted_results = ""
        for result in results:
            formatted_results += f"Title: {result['title']}\n"
            formatted_results += f"Content: {result['body']}\n"
            formatted_results += f"Source: {result['href']}\n\n"

        return formatted_results
    except Exception as e:
        return f"Search error: {str(e)}"



from textual.app import ComposeResult
from textual.widgets import (
    Label,
    Static,
)


class ChatMessage(Static):
    """A single chat message widget with label"""

    def __init__(self, content: str, role: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.role: str = role
        self.content: str = content
        self.add_class(f"message-{role}")

    def compose(self) -> ComposeResult:
        if self.role == "user":
            label_text = "User:"
        elif self.role == "web_search":
            label_text = "Web Search Results:"
        elif self.role == "document_search":
            label_text = "Vector Search Results:"
        elif self.role == "memory_search":
            label_text = "Memory Search Results:"
        else:
            label_text = "Assistant:"
        yield Label(label_text, classes=f"message-label-{self.role}")
        yield Static(self.content, classes="message-content")

    def update_content(self, new_content: str) -> None:
        """Update the message content"""
        self.content = new_content
        content_widget = self.query_one(".message-content", Static)
        content_widget.update(new_content)



from textual.theme import Theme

open_terminalui_theme = Theme(
    name="open_terminalui",
    primary="lime",
    secondary="white",
    accent="lime",
    foreground="#ffffff",
    background="#000000",
    success="green",
    warning="yellow",
    error="red",
    surface="#111111",
    panel="#444444",
    dark=True,
    variables={
        "footer-key-foreground": "lime",
    },
)



import hashlib
import os
from pathlib import Path
from typing import List, Tuple

import chromadb
from pypdf import PdfReader


class DocumentManager:
    """Manages PDF documents and their vector embeddings using ChromaDB"""

    def __init__(self, storage_path: str | None = None):
        """Initialize the document manager with ChromaDB client"""
        if storage_path is None:
            # Default to ~/.open-terminalui/chroma_db
            app_dir = Path.home() / ".open-terminalui"
            app_dir.mkdir(exist_ok=True)
            storage_path = str(app_dir / "chroma_db")

        self.storage_path = storage_path
        self.client = chromadb.PersistentClient(path=storage_path)

        # Get or create the documents collection
        self.collection = self.client.get_or_create_collection(
            name="documents",
            metadata={"description": "PDF document chunks with embeddings"},
        )

    def _extract_text_from_pdf(self, file_path: str) -> str:
        """Extract text content from a PDF file"""
        try:
            reader = PdfReader(file_path)
            text = ""
            for page in reader.pages:
                text += page.extract_text() + "\n"
            return text
        except Exception as e:
            raise ValueError(f"Failed to extract text from PDF: {str(e)}")

    def _chunk_text(
        self, text: str, chunk_size: int = 500, overlap: int = 50
    ) -> List[str]:
        """Split text into overlapping chunks"""
        words = text.split()
        chunks = []

        for i in range(0, len(words), chunk_size - overlap):
            chunk = " ".join(words[i : i + chunk_size])
            if chunk:
                chunks.append(chunk)

        return chunks

    def _get_file_hash(self, file_path: str) -> str:
        """Generate a hash for the file to use as unique identifier"""
        return hashlib.md5(file_path.encode()).hexdigest()

    def add_document(self, file_path: str) -> Tuple[bool, str]:
        """
        Add a PDF document to the vector store

        Returns:
            Tuple of (success: bool, message: str)
        """
        # Validate file exists and is a PDF
        if not os.path.exists(file_path):
            return False, f"File not found: {file_path}"

        if not file_path.lower().endswith(".pdf"):
            return False, "Only PDF files are supported"

        try:
            # Check if document already exists
            file_hash = self._get_file_hash(file_path)
            existing = self.collection.get(where={"file_hash": file_hash})

            if existing and existing["ids"]:
                return False, f"Document already exists: {os.path.basename(file_path)}"

            # Extract text from PDF
            text = self._extract_text_from_pdf(file_path)

            if not text.strip():
                return False, "PDF appears to be empty or contains no extractable text"

            # Chunk the text
            chunks = self._chunk_text(text)

            # Generate IDs and metadata for each chunk
            file_name = os.path.basename(file_path)
            ids = [f"{file_hash}_chunk_{i}" for i in range(len(chunks))]
            metadatas = [
                {
                    "file_path": file_path,
                    "file_name": file_name,
                    "file_hash": file_hash,
                    "chunk_index": i,
                    "total_chunks": len(chunks),
                }
                for i in range(len(chunks))
            ]

            # Add to ChromaDB (it will automatically generate embeddings)
            self.collection.add(
                documents=chunks,
                ids=ids,
                metadatas=metadatas,
            )

            return True, f"Successfully added {len(chunks)} chunks from {file_name}"

        except Exception as e:
            return False, f"Error adding document: {str(e)}"

    def remove_document(self, file_path: str) -> Tuple[bool, str]:
        """
        Remove a document from the vector store

        Returns:
            Tuple of (success: bool, message: str)
        """
        try:
            file_hash = self._get_file_hash(file_path)

            # Get all chunks for this document
            results = self.collection.get(where={"file_hash": file_hash})

            if not results["ids"]:
                return False, f"Document not found: {os.path.basename(file_path)}"

            # Delete all chunks
            self.collection.delete(ids=results["ids"])

            return True, f"Successfully removed {os.path.basename(file_path)}"

        except Exception as e:
            return False, f"Error removing document: {str(e)}"

    def list_documents(self) -> List[Tuple[str, str, int]]:
        """
        List all documents in the vector store

        Returns:
            List of tuples: (file_path, file_name, num_chunks)
        """
        try:
            # Get all documents
            results = self.collection.get()

            if not results["metadatas"]:
                return []

            # Group by file_hash to get unique documents
            docs_by_hash = {}
            for metadata in results["metadatas"]:
                file_hash = metadata["file_hash"]
                if file_hash not in docs_by_hash:
                    docs_by_hash[file_hash] = {
                        "file_path": metadata["file_path"],
                        "file_name": metadata["file_name"],
                        "chunk_count": 0,
                    }
                docs_by_hash[file_hash]["chunk_count"] += 1

            # Convert to list of tuples
            return [
                (doc["file_path"], doc["file_name"], doc["chunk_count"])
                for doc in docs_by_hash.values()
            ]

        except Exception as _:
            return []

    def search_documents(
        self, query: str, top_k: int = 5
    ) -> List[Tuple[str, str, float]]:
        """
        Search for relevant document chunks using vector similarity

        Args:
            query: The search query
            top_k: Number of top results to return

        Returns:
            List of tuples: (chunk_text, file_name, similarity_score)
        """
        try:
            results = self.collection.query(
                query_texts=[query],
                n_results=top_k,
            )

            if not results["documents"] or not results["documents"][0]:
                return []

            chunks = []
            documents = results["documents"][0]
            metadatas = results["metadatas"][0] if results["metadatas"] else []
            distances = results["distances"][0] if results["distances"] else []

            for i, doc in enumerate(documents):
                if i < len(metadatas):
                    metadata = metadatas[i]
                    # ChromaDB returns distances, lower is better
                    # Convert to similarity score (1 - normalized_distance)
                    distance = distances[i] if i < len(distances) else 0
                    similarity = max(0, 1 - distance)

                    chunks.append((doc, metadata["file_name"], similarity))

            return chunks

        except Exception as _:
            return []



import ollama
from textual import on, work
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.widgets import (
    Footer,
    Header,
    Input,
    Label,
    ListView,
    Static,
    Switch,
)

from open_terminalui._models import Chat, Message
from open_terminalui._themes import open_terminalui_theme
from open_terminalui.chat_manager import ChatManager
from open_terminalui.components import ChatListItem, ChatMessage
from open_terminalui.document_manager import DocumentManager
from open_terminalui.memory_manager import MemoryManager
from open_terminalui.screens.document_screen import DocumentManagerScreen
from open_terminalui.tools import document_search, memory_search, web_search


class OpenTerminalUI(App):
    CSS_PATH = "styles.tcss"
    BINDINGS = [
        ("ctrl+n", "new_chat", "New Chat"),
        ("ctrl+b", "toggle_sidebar", "Toggle Sidebar"),
        ("ctrl+d", "delete_chat", "Delete Chat"),
        ("ctrl+k", "manage_documents", "Manage Documents"),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.chat_history = []
        self.current_assistant_message: ChatMessage
        self.chat_manager = ChatManager()
        self.doc_manager = DocumentManager()
        self.memory_manager = MemoryManager()
        self.current_chat: Chat
        self.sidebar_visible = True

    def compose(self) -> ComposeResult:
        yield Header()
        with Horizontal(id="main_container"):
            with Vertical(id="sidebar"):
                yield Label("Chat History", id="sidebar_title")
                yield ListView(id="chat_list")
            with Vertical(id="container"):
                with VerticalScroll(id="chat_container"):
                    pass  # Messages will be added dynamically
                yield Static(" ", id="loading_indicator")
                with Vertical(id="input_bar"):
                    yield Input(
                        type="text", id="input", placeholder="Type a message..."
                    )
                    with Horizontal(id="buttons_container"):
                        with Horizontal(classes="toggle_container"):
                            yield Label("Search:", classes="toggle_label")
                            yield Switch(value=False, id="search_switch")
                        with Horizontal(classes="toggle_container"):
                            yield Label("Documents:", classes="toggle_label")
                            yield Switch(value=False, id="documents_switch")
                        with Horizontal(classes="toggle_container"):
                            yield Label("Memory:", classes="toggle_label")
                            yield Switch(value=False, id="memory_switch")
                        with Horizontal(classes="toggle_container"):
                            yield Label("Logs:", classes="toggle_label")
                            yield Switch(value=False, id="logs_switch")
        yield Footer()

    def on_mount(self) -> None:
        """Initialize a new chat on startup"""
        self.register_theme(open_terminalui_theme)
        self.theme = "open_terminalui"
        self._refresh_chat_list()
        self._new_chat()

    @work(exclusive=True, thread=True)
    def stream_ollama_response(
        self,
        content: str,
        use_search: bool = False,
        use_logs: bool = False,
        use_documents: bool = False,
        use_memory: bool = False,
    ) -> None:
        # Select and update loading indicator
        loading_indicator = self.query_one("#loading_indicator", Static)

        # If search is enabled, perform web search first
        messages_to_send = self.chat_history.copy()
        web_search_results = None
        if use_search:
            self.call_from_thread(loading_indicator.update, "Searching the web...")
            web_search_results = web_search(query=content)

            # Add search results as system context
            messages_to_send = [
                {
                    "role": "system",
                    "content": f"Use the following web search results to help answer the user's question:\n\n{web_search_results}",
                }
            ] + messages_to_send

            # Always save logs to database
            log_message_data = Message(role="web_search", content=web_search_results)
            self.current_chat.messages.append(log_message_data)

            # Only display in UI if logs switch is enabled
            if use_logs:
                chat_container = self.query_one("#chat_container", VerticalScroll)
                log_message = ChatMessage(web_search_results, "web_search")
                self.call_from_thread(chat_container.mount, log_message)
                self.call_from_thread(chat_container.scroll_end, animate=False)

        # If documents is enabled, perform vector search
        document_search_results = None
        if use_documents:
            self.call_from_thread(
                loading_indicator.update, "Searching vector database..."
            )
            document_search_results = document_search(
                doc_manager=self.doc_manager, query=content
            )

            # Add search results as system context
            messages_to_send = [
                {
                    "role": "system",
                    "content": f"Use the following vector search results to help answer the user's question:\n\n{document_search_results}",
                }
            ] + messages_to_send

            # Always save logs to database
            log_message_data = Message(
                role="document_search", content=document_search_results
            )
            self.current_chat.messages.append(log_message_data)

            # Only display in UI if logs switch is enabled
            if use_logs:
                chat_container = self.query_one("#chat_container", VerticalScroll)
                log_message = ChatMessage(document_search_results, "document_search")
                self.call_from_thread(chat_container.mount, log_message)
                self.call_from_thread(chat_container.scroll_end, animate=False)

        # If memory is enabled, perform memory search
        memory_search_results = None
        if use_memory:
            self.call_from_thread(loading_indicator.update, "Searching chat memory...")
            memory_search_results = memory_search(
                memory_manager=self.memory_manager, query=content
            )

            # Add search results as system context
            messages_to_send = [
                {
                    "role": "system",
                    "content": f"Use the following chat summaries from other messages to help answer the user's question:\n\n{memory_search_results}",
                }
            ] + messages_to_send

            # Always save logs to database
            log_message_data = Message(
                role="memory_search", content=memory_search_results
            )
            self.current_chat.messages.append(log_message_data)

            # Only display in UI if logs switch is enabled
            if use_logs:
                chat_container = self.query_one("#chat_container", VerticalScroll)
                log_message = ChatMessage(memory_search_results, "memory_search")
                self.call_from_thread(chat_container.mount, log_message)
                self.call_from_thread(chat_container.scroll_end, animate=False)

        self.call_from_thread(loading_indicator.update, "Thinking...")

        # Select chat history widget
        chat_container = self.query_one("#chat_container", VerticalScroll)

        # Stream ollama response
        stream = ollama.chat(model="llama3.2", messages=messages_to_send, stream=True)
        accumulated_text = ""

        for i, chunk in enumerate(stream):
            accumulated_text += chunk["message"]["content"]

            # Create assistant message widget on first chunk
            if i == 0:
                self.chat_history.append(
                    {"role": "assistant", "content": accumulated_text}
                )
                assistant_message = Message(role="assistant", content=accumulated_text)
                self.current_chat.messages.append(assistant_message)

                self.current_assistant_message = ChatMessage(
                    accumulated_text, "assistant"
                )
                self.call_from_thread(
                    chat_container.mount, self.current_assistant_message
                )
                self.call_from_thread(loading_indicator.update, " ")
            # Update existing assistant message
            else:
                self.chat_history[-1]["content"] = accumulated_text
                self.current_chat.messages[-1].content = accumulated_text
                self.call_from_thread(
                    self.current_assistant_message.update_content, accumulated_text
                )

            # Auto-scroll to bottom
            self.call_from_thread(chat_container.scroll_end, animate=False)

        # Save chat to database after streaming completes
        self.chat_manager.save_chat(self.current_chat)

        # Save chat to vector database too
        self.memory_manager.save_chat(self.current_chat)

        # Refresh sidebar to update chat title/timestamp
        self.call_from_thread(self._refresh_chat_list)

    # ---------- Private Methods ----------
    def _refresh_chat_list(self) -> None:
        """Refresh the sidebar chat list"""
        chat_list = self.query_one("#chat_list", ListView)
        chat_list.clear()

        chats = self.chat_manager.list_chats()
        for chat in chats:
            if chat.id is not None:
                chat_list.append(ChatListItem(chat.id, chat.title))

    def _new_chat(self) -> None:
        """Create a new chat and clear the UI (not saved to DB until it has messages)"""
        self.current_chat = Chat.create_unsaved()
        self.chat_history = []

        # Clear chat container
        chat_container = self.query_one("#chat_container", VerticalScroll)
        chat_container.remove_children()

    def _load_chat(self, chat_id: int) -> None:
        """Load an existing chat"""
        chat = self.chat_manager.load_chat(chat_id)
        if chat is None:
            return

        self.current_chat = chat
        self.chat_history = chat.to_ollama_messages()

        # Clear and reload chat messages in UI
        chat_container = self.query_one("#chat_container", VerticalScroll)
        chat_container.remove_children()

        # Check if logs should be displayed
        logs_switch = self.query_one("#logs_switch", Switch)
        show_logs = logs_switch.value

        for message in chat.messages:
            # Skip log messages if logs switch is off
            if (
                message.role in ("web_search", "document_search", "memory_search")
                and not show_logs
            ):
                continue

            msg_widget = ChatMessage(message.content, message.role)
            chat_container.mount(msg_widget)

        chat_container.scroll_end(animate=False)

    # ---------- Handlers ----------

    @on(Input.Submitted, "#input")
    def handle_input_submission(self) -> None:
        input_widget = self.query_one("#input", Input)
        content: str = input_widget.value.strip()

        if not content:
            return

        # Check if search, logs, and documents are enabled
        search_switch = self.query_one("#search_switch", Switch)
        logs_switch = self.query_one("#logs_switch", Switch)
        documents_switch = self.query_one("#documents_switch", Switch)
        memory_switch = self.query_one("#memory_switch", Switch)
        use_search = search_switch.value
        use_logs = logs_switch.value
        use_documents = documents_switch.value
        use_memory = memory_switch.value

        # Add user message to chat
        user_message = Message(role="user", content=content)
        self.current_chat.messages.append(user_message)
        self.chat_history.append({"role": "user", "content": content})

        # Create and mount user message widget
        chat_container = self.query_one("#chat_container", VerticalScroll)
        user_msg = ChatMessage(content, "user")
        chat_container.mount(user_msg)
        chat_container.scroll_end(animate=False)

        input_widget.clear()
        self.stream_ollama_response(
            content, use_search, use_logs, use_documents, use_memory
        )

    @on(ListView.Selected, "#chat_list")
    def handle_chat_selection(self, event: ListView.Selected) -> None:
        """Handle chat selection from sidebar"""
        if isinstance(event.item, ChatListItem):
            self._load_chat(event.item.chat_id)

    @on(Switch.Changed, "#logs_switch")
    def handle_logs_toggle(self, event: Switch.Changed) -> None:
        """Handle logs switch toggle - reload current chat to show/hide logs"""
        if self.current_chat and self.current_chat.id is not None:
            self._load_chat(self.current_chat.id)

    # ---------- Actions ----------

    def action_new_chat(self) -> None:
        """Action to create a new chat"""
        self._new_chat()

    def action_delete_chat(self) -> None:
        """Delete the currently selected chat from the sidebar"""
        chat_list = self.query_one("#chat_list", ListView)

        # Get the currently highlighted item
        if chat_list.index is None:
            return

        selected_item = chat_list.highlighted_child
        if not isinstance(selected_item, ChatListItem):
            return

        chat_id_to_delete = selected_item.chat_id

        # Delete from database
        self.chat_manager.delete_chat(chat_id_to_delete)
        self.memory_manager.delete_chat(chat_id_to_delete)

        # If we're deleting the current chat, create a new one
        if (
            self.current_chat
            and self.current_chat.id is not None
            and self.current_chat.id == chat_id_to_delete
        ):
            self._new_chat()

        # Refresh the sidebar
        self._refresh_chat_list()

    def action_toggle_sidebar(self) -> None:
        """Toggle sidebar visibility"""
        sidebar = self.query_one("#sidebar")
        sidebar.display = not sidebar.display
        self.sidebar_visible = sidebar.display

    def action_manage_documents(self) -> None:
        """Open the document management screen"""
        self.push_screen(DocumentManagerScreen(self.doc_manager))
